library markdown.src.document;

import 'ast.dart';
import 'block_parser.dart';
import 'extension_set.dart';
import 'inline_parser.dart';

typedef BlockParser _BlockParserBuilder(List<String> lines, Document document);
typedef InlineParser _InlineParserBuilder(String text, Document document);

/// Maintains the context needed to parse a Markdown document.
class Document {
  final Map<String, Link> refLinks = {};
  Iterable<BlockSyntax> blockSyntaxes;
  Iterable<InlineSyntax> inlineSyntaxes;
  ExtensionSet extensionSet;
  Resolver linkResolver;
  Resolver imageLinkResolver;
  ///An application specific instance.
  final options;
  ///Encapsultes the instantiation of parsers for easy customization (tom)
  final _BlockParserBuilder _blockParserBuilder;
  final _InlineParserBuilder _inlineParserBuilder;

  Document(
      {Iterable<BlockSyntax> blockSyntaxes,
      Iterable<InlineSyntax> inlineSyntaxes,
      ExtensionSet extensionSet,
      this.linkResolver,
      this.imageLinkResolver,
      _BlockParserBuilder blockParserBuilder: _newBlockParser,
      _InlineParserBuilder inlineParserBuilder: _newInlineParser,
      this.options}):
      _blockParserBuilder = blockParserBuilder,
      _inlineParserBuilder = inlineParserBuilder {
    this.extensionSet = extensionSet ?? ExtensionSet.commonMark;
    this.blockSyntaxes = new Set()
      ..addAll(blockSyntaxes ?? [])
      ..addAll(this.extensionSet.blockSyntaxes);
    this.inlineSyntaxes = new Set()
      ..addAll(inlineSyntaxes ?? [])
      ..addAll(this.extensionSet.inlineSyntaxes);
  }
  ///A lightweight, full-customized document.
  Document.plain(
      this._blockParserBuilder, this._inlineParserBuilder, {this.options});

  BlockParser getBlockParser(List<String> lines)
  => _blockParserBuilder(lines, this);
  InlineParser getInlineParser(String text)
  => _inlineParserBuilder(text, this);

  /// Parses the given [lines] of Markdown to a series of AST nodes.
  List<Node> parseLines(List<String> lines) {
    List<Node> nodes = getBlockParser(lines).parseLines();
    _parseInlineContent(nodes);
    return nodes;
  }

  /// Parses the given inline Markdown [text] to a series of AST nodes.
  List<Node> parseInline(String text) => getInlineParser(text).parse();

  void _parseInlineContent(List<Node> nodes) {
    for (int i = 0; i < nodes.length; i++) {
      var node = nodes[i];
      if (node is UnparsedContent) {
        List<Node> inlineNodes = parseInline(node.textContent);
        nodes.removeAt(i);
        nodes.insertAll(i, inlineNodes);
        i += inlineNodes.length - 1;
      } else if (node is Element && node.children != null) {
        _parseInlineContent(node.children);
      }
    }
  }
}

BlockParser _newBlockParser(List<String> lines, Document document)
=> new BlockParser(lines, document);
InlineParser _newInlineParser(String text, Document document)
=> new InlineParser(text, document);


class Link {
  final String id;
  final String url;
  final String title;
  Link(this.id, this.url, this.title);
}
